/*******************************************************************************
 * Copyright (c) 2009 Mia-Software.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Nicolas Bros (Mia-Software) - initial API and implementation
 *    
 *******************************************************************************/

package org.eclipse.gmt.modisco.common.editor.editors;

import org.eclipse.gmt.modisco.common.editor.editors.EditorConfiguration.MetaclassesSortMode;
import org.eclipse.gmt.modisco.common.editor.util.ImageProvider;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.ToolBar;
import org.eclipse.swt.widgets.ToolItem;

public class MetaclassViewToolBar {
	private final ToolBar toolBar;
	private final MetaclassViewer metaclassViewer;
	private ToolItem btnSortByName;
	private ToolItem btnSortByCount;
	private ToolItem btnShowMetaclassesFullQualifiedName;
	private ToolItem btnShowEmptyMetaclasses;
	private ToolItem btnGroupByPackage;
	private ToolItem btnDisplayInstancesOfSubclasses;
	private ToolItem btnDerivationTree;
	
	private final EditorConfiguration editorConfiguration;

	public MetaclassViewToolBar(Composite parent, MetaclassViewer metaclassViewer, EditorConfiguration editorConfiguration) {
		this.editorConfiguration = editorConfiguration;
		GridLayout gridLayout = new GridLayout(2, false);
		gridLayout.horizontalSpacing = 30;
		// the margins are set so that the metaclass pane is aligned with the tree pane
		gridLayout.marginHeight = 2;
		gridLayout.marginTop = 1;
		parent.setLayout(gridLayout);
		
		this.metaclassViewer = metaclassViewer;
		toolBar = new ToolBar(parent, SWT.NONE);
		GridData toolBarGridData = new GridData();
		toolBarGridData.grabExcessHorizontalSpace = false;
		toolBarGridData.horizontalAlignment = SWT.FILL;
		toolBar.setLayoutData(toolBarGridData);

		createToolButtons();
		initialize();
	}

	private void createToolButtons() {
		btnSortByName = new ToolItem(toolBar, SWT.RADIO);
		btnSortByName.setToolTipText("Sort by Name");
		btnSortByName.setImage(ImageProvider.getInstance().getSortInstancesIcon());
		btnSortByName.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				metaclassViewer.setSortMode(EditorConfiguration.MetaclassesSortMode.ByName);
			}
		});
		
		btnSortByCount = new ToolItem(toolBar, SWT.RADIO);
		btnSortByCount.setToolTipText("Sort by Count");
		btnSortByCount.setImage(ImageProvider.getInstance().getSortNumericallyIcon());
		btnSortByCount.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				metaclassViewer.setSortMode(EditorConfiguration.MetaclassesSortMode.ByCount);
			}
		});
		
		btnShowMetaclassesFullQualifiedName = new ToolItem(toolBar, SWT.CHECK);
		btnShowMetaclassesFullQualifiedName.setToolTipText("Show Full Qualified Names");
		btnShowMetaclassesFullQualifiedName.setImage(ImageProvider.getInstance().getShowFullQualifiedNamesIcon());
		btnShowMetaclassesFullQualifiedName.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				metaclassViewer.setShowMetaclassesFullQualifiedNames(btnShowMetaclassesFullQualifiedName.getSelection());
			}
		});

		btnShowEmptyMetaclasses = new ToolItem(toolBar, SWT.CHECK);
		btnShowEmptyMetaclasses.setToolTipText("Show Empty Metaclasses");
		btnShowEmptyMetaclasses.setImage(ImageProvider.getInstance().getShowEmptyMetaclassesIcon());
		btnShowEmptyMetaclasses.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				metaclassViewer.setShowEmptyMetaclasses(btnShowEmptyMetaclasses.getSelection());
			}
		});

		btnGroupByPackage = new ToolItem(toolBar, SWT.CHECK);
		btnGroupByPackage.setToolTipText("Group By Package");
		btnGroupByPackage.setImage(ImageProvider.getInstance().getPackageIcon());
		btnGroupByPackage.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				metaclassViewer.setGroupByPackage(btnGroupByPackage.getSelection());
			}
		});

		btnDisplayInstancesOfSubclasses = new ToolItem(toolBar, SWT.CHECK);
		btnDisplayInstancesOfSubclasses.setToolTipText("Display Instances of Subclasses");
		btnDisplayInstancesOfSubclasses.setImage(ImageProvider.getInstance()
				.getDerivedInstancesIcon());
		btnDisplayInstancesOfSubclasses.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				metaclassViewer.setDisplayInstancesOfSubclasses(btnDisplayInstancesOfSubclasses
						.getSelection());
			}
		});

		btnDerivationTree = new ToolItem(toolBar, SWT.CHECK);
		btnDerivationTree.setToolTipText("Show Derivation Tree");
		btnDerivationTree.setImage(ImageProvider.getInstance()
				.getDerivationTreeIcon());
		btnDerivationTree.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				metaclassViewer.setShowDerivationTree(btnDerivationTree
						.getSelection());
				btnGroupByPackage.setEnabled(!btnDerivationTree.getSelection());
			}
		});
	}
	
	/** Initialize the tool bar buttons with their initial value*/
	private void initialize() {
		MetaclassesSortMode metaclassesSortMode = this.editorConfiguration.getMetaclassesSortMode();
		if(metaclassesSortMode == MetaclassesSortMode.ByName) {
			btnSortByName.setSelection(true);
		} else if(metaclassesSortMode == MetaclassesSortMode.ByCount) {
			btnSortByCount.setSelection(true);
		}
		btnShowMetaclassesFullQualifiedName.setSelection(this.editorConfiguration.isShowMetaclassesFullQualifiedNames());
		btnShowEmptyMetaclasses.setSelection(this.editorConfiguration.isShowEmptyMetaclasses());
		btnGroupByPackage.setSelection(this.editorConfiguration.isGroupByPackage());
		btnDisplayInstancesOfSubclasses.setSelection(this.editorConfiguration.isDisplayInstancesOfSubclasses());
		btnDerivationTree.setSelection(this.editorConfiguration.isShowDerivationTree());
		
		btnGroupByPackage.setEnabled(!this.editorConfiguration.isShowDerivationTree());
	}	
}
